/*
 * Home Automation MODBUS
 *
 * 2015-10-12, 2015-12-30, 2016-01-06 Kouichi Hirabayashi
 * 1 WMS-PE6N-00A007
 * 2 WMS-PE6N-00A007
 * 3 WMS-PE6N-00A007
 * 4 WMS-PE6N-00A007
 * 5 WMB-DIO8R-0D000
 * 6 WMB-AI8-15D000
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <termios.h>
#include <netinet/in.h>
#include "circuit.h"

int serOpen(char *, int);
int serClose(int);
int check(int);
int SetPar();
int GetPower(int, int, int);
int GetEnergy(int, int, int);
int diag(int);
int GetCurrent(int, int, int);
int GetVoltage(int, int, int);
int GetFrequency(int, int, int);
int RedPar(int);
int GetPar(int);
int getAin(int, int, double*);
int getCount(int, int);
int getOnoff(int, int);
int resetCounter(int, int, int);
int getResponse(int);
int getResponse2(int);
int Red(int, int, int);
int recv(int, char*, int);
int Energy();
int Power();
int Voltage();
int Current();
int Frequency();
int ResetMinMax(int, char**);
int ResetEnergy(int, char**);
int Circuit2IdCh(int, int*, int*, int*);
int dump(char*, int);
int Module();
void Usage();
void error(char *, int);
int Wrt(int, int, int, ...);

#ifdef DIRECT
#define PORT	"/dev/cuau0"
#else
#define HOST	"192.168.3.40"	// NPort5450#1 19200,8,1,No RS-485 Wire
#define PORT	4002
#endif

unsigned short Crc(unsigned char*, int);
unsigned short Lrc(unsigned char*, int);

struct data {
  double i[9];	// cuurent
  double v[9];	// voltage
  double p[9];	// power
  double e[9];	// energy
  double f[9];	// frequency
} data[50];

int fd;
int Cflg;	// counter reset
int Iflg;	// illuminance
int Mflg;	// reset min max
int Rflg;	// reset energy data
int Sflg;	// set parameters
int Wflg;	// water, gas
int aflg;	// read analog input
int cflg;	// read counter
int dflg;	// diagnosis
int eflg;	// energy
int fflg;	// frequency
int iflg;	// current
int mflg;	// check module
int pflg;	// power
int vflg;	// voltage
int sflg;	// read settings
int xflg;	// debug
char *cmd;
char buf[BUFSIZ];

int main(argc, argv) char **argv;
{
  char *s;
  int i, j;
  double val;

  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
	for (s = argv[0]+1; *s; s++)
		switch (*s) {
		case 'C':
			Cflg++;
			break;
		case 'I':
			Iflg++;
			break;
		case 'M':
			Mflg++;
			break;
		case 'R':
			Rflg++;
			break;
		case 'S':
			Sflg++;
			break;
		case 'W':
			Wflg++;
			break;
		case 'a':
			aflg++;
			break;
		case 'c':
			cflg++;
			break;
		case 'd':
			dflg++;
			break;
		case 'e':
			eflg++;
			break;
		case 'f':
			fflg++;
			break;
		case 'i':
			iflg++;
			break;
		case 'm':
			mflg++;
			break;
		case 'p':
			pflg++;
			break;
		case 's':
			sflg++;
			break;
		case 'v':
			vflg++;
			break;
		case 'x':
			xflg++;
			break;
		default:
			Usage();
		}
  if ((fd = serOpen(HOST, PORT)) < 0)
	error("can not open %s", (int)HOST);
  usleep(50000);


  if (Cflg) {
	int ch = 0, v = 0;

	if (argc < 1 || 2 < argc)
		Usage();
	if (0 < argc)
		ch = atoi(argv[0]);
	if (1 < argc)
		v = atoi(argv[1]);
	resetCounter(fd, ch, v);
  }
  if (Mflg)
	ResetMinMax(argc, argv);
  if (Rflg)
	ResetEnergy(argc, argv);
  if (Iflg) {	// illuminance
	getAin(fd, 1, &val);
	printf("%.0f\n", val * 7000.0 / 5.0);
  }
  else if (aflg) {
	int i;
	for (i = 1; i <= 8; i++) {
		getAin(fd, i, &val);
		printf("CH %d %.3f V\n", i, val);
	}
  }
  else if (Wflg) {	//water, gas
	int g, w;
	w = getCount(fd, 1);
	w += getCount(fd, 2);
	g = getCount(fd, 3);
	g += getCount(fd, 4);
	printf("%d %.2f\n", w, (double)g * 0.05);
  }
  if (cflg) {
	if (argc > 0) {
		for (i = 0; i < argc; i++)
			printf("CH%d %d %d\n", atoi(argv[i]),
				getCount(fd, atoi(argv[i])),
				getOnoff(fd, atoi(argv[i])));
	}
	else
		for (i = 1; i <= 8; i++)
			printf("CH%d %d %d\n", i, getCount(fd, i),
				getOnoff(fd, i));

  }
  if (dflg)
	for (i = 1; i <= 4; i++)
		diag(i);
  if (eflg)
	Energy();
  if (fflg)
	Frequency();
  if (iflg)
	Current();
  if (mflg)
	Module();
  if (pflg)
	Power();
  if (vflg)
	Voltage();
  if (Sflg)
	SetPar();
  if (sflg)
	for (i = 1; i <= 4; i++)
		GetPar(i);
  serClose(fd);
}

int check(id)
{
  char *s;

  Red(id, 0x07D0, 20);
  s = buf + 3;
	  printf("%.20s\n", s);
  return 0;
}

/* ѥ᡼񤭹 (par[] ) */

int SetPar()
{
  static struct { char * typ; int adr, nval, val[6], check, max; } par[] = {
	{ "ID", 0, 1, { 0, 0, 0, 0, 0, 0 }, 0, 4 },
	{ "ʬ", 0x1000, 6, { 0, 0, 0, 0, 0, 0}, 0, 5 },
	{ "CHѾ", 0x1017, 1, { 0, 0, 0, 0, 0, 0 }, 0, 0xfff },
	{ "CHŰ", 0x1016, 1, { 0, 0, 0, 0, 0, 0 }, 0, 0x3f },
	{ "Ű", 0x1006, 2, { 0, 0, 0, 0, 0, 0 }, 0, 1 },
	{ "ʰ켡Ű", 0x1014, 2, { 0, 0, 0, 0, 0, 0 }, 0, 0x12 },
	{ "ή", 0x1008, 6, { 0, 0, 0, 0, 0, 0 }, 0, 5 },
	{ "ʰ켡ή", 0x100E, 6, { 0, 0, 0, 0, 0, 0 }, 0, 0x27 },
	{ NULL, 0, 0, { 0, 0, 0, 0, 0, 0 }, 0, 0 }
  };
	
  FILE *fp;
  char *s, *t, c, type[128];
  int i, j, k, n, id, adr, v[6];

  if ((fp = fopen("PE6N.set", "r")) == NULL)
	error("Can;t open %s", (int)"PE6N.set");
  k = 0;
  while (fgets(buf, sizeof(buf), fp) != NULL) {
	if (*buf == '#' || *buf == '\n')
		continue;
	*type = '\0';
	for (i = 0; i < 6; i++)
		v[i] = -1;
	n = sscanf(buf, "%s %x %x %x %x %x %x", type, &v[0], &v[1],
		&v[2], &v[3], &v[4], &v[5]);
	for (i = 0; par[i].typ != NULL; i++) {
		if (strstr(buf, par[i].typ) != NULL)
			break;
	}
	if (par[i].typ == NULL)
		error("Unknown type: %s", (int)buf);
	if (i == 0) {
		k = 0;
		par[0].check = 1;
		for (j = 1; j < 8; j++)
			par[j].check = 0;
	}
	else {
		k++;
		par[i].check++;
	}
	if (n != par[i].nval + 1) {
		fprintf(stderr, "record: %s", buf);
		error("# of parameters must be %d", par[i].nval);
	}
	for (j = 0; j < par[i].nval; j++) {
		if (par[i].max < v[j] || v[j] < 0) {
			fprintf(stderr, "%s", buf);
			error("llegal parameter value: 0x%02x", v[j]);
		}
		par[i].val[j] = v[j];
	}
	if (k < 7) {
		continue;
	}
	// output
	for (j = 0; par[j].typ != NULL; j++) {
		if (par[j].check == 1)
			continue;
		if (par[j].check < 1)
			fprintf(stderr, "ID %d: missing %s\n", par[0].val[0],
				par[j].typ);
		else
			fprintf(stderr, "ID %d: duplicate %s\n", par[0].val[0],
				par[j].typ);
		error("Parameter error.", 0);
		printf("ID(%d) check(%d)\n", par[0].val[0], par[j].check);
	}
	for (i = 1; par[i].typ != NULL; i++) {
		int adr = par[i].adr;
		for (j = 0; j < par[i].nval; j++, adr++) {
			printf("id(%d) adr(0x%04x) 1 0x%x\n", par[0].val[0], adr, par[i].val[j]);
			Wrt(par[0].val[0], adr, 1, par[i].val[j]);
			usleep(100000);
		}
	}
  }
  fclose(fp);
  return 0;
}

static char *err[] = {
  "", "ޥ", "ɥ쥹", "ǡ"
};

int GetPower(id, ch, ab)
{
  int i, k, adr;
  long long int d;
  unsigned long h, l;
  short e;
  char *s, *t;
  double v;

  // power
  adr = 0x0380 + (ch - 1) * 0x36 + ab * 0x1B;
  k = (id - 1) * 12 + (ch - 1) + ab * 6;	// circuittab[] index
  if (xflg)
	  printf("ID %d  CH%d-%s (0x%04x, 0x%04x) %s\n", id, ch, ab ? "B" : "A", adr,
		0x0388 + (ch - 1) * 0x36 + ab * 0x1B, circuittab[k].s);
  if (adr < 0x0380 || 0x04C7 < adr) {
	fprintf(stderr, "GetPower(%d, %d, %d): illegal address.\n",
		id, ch, ab);
	return -1;
  }
  Red(id, adr, 27);
  s = buf + 3;
  for (i = 0; i < 9; i++) {
	if ((i % 3) != 2) {
		t = (char *)&d;
		*t++ = *(s + 7);
		*t++ = *(s + 6);
		*t++ = *(s + 5);
		*t++ = *(s + 4);
		*t++ = *(s + 3);
		*t++ = *(s + 2);
		*t++ = *(s + 1);
		*t++ = *(s + 0);
		if ((s[0] & 0xff) == 0x80)
			v = 0.0;
		else
			v = (double)d * 1e-2;
		if (xflg)
			printf("%d: %.2f\n", i, v);
		data[k].p[i] = v;
		if (xflg)
			printf("i(%d) p(%.2f)\n", i, v);
		s += 8;
	}
	else {
		e = ntohs(*(short *)s);
		v = (double)e * 1e-3;
		if (xflg)
			printf("%d: %.3f\n", i, v);
		if ((s[0] & 0xff) == 0x80)
				v = 0.0;
		else
			data[k].p[i] = v;
		if (xflg)
			printf("i(%d) pf(%.3f)\n", i, v);
		s+= 2;
	}
  }
  return 0;
}

int GetEnergy(id, ch, ab)
{
  int i, k, adr;
  unsigned long l, h;
  unsigned long long d;
  char *s, *t;
  double v;

  adr = 0x0500 + (ch - 1) * 0x30 + ab * 0x18;
  k = (id - 1) * 12 + (ch - 1) + ab * 6;	// circuittab[] index
  if (xflg)
	printf("ID %d  CH%d-%s (0x%04x) %s\n", id, ch, ab ? "B" : "A", adr,
		circuittab[k].s);
  if (adr < 0x0500 || 0x061C < adr) {
	fprintf(stderr, "GetEnergy(%d, %d, %d): illegal address.\n",
		id, ch, ab);
	return -1;
  }
  Red(id, adr, 48);
  s = buf + 3;
  for (i = 0; i < 6; i++) {
	t = (char *)&d;
	*t++ = *(s + 7);
	*t++ = *(s + 6);
	*t++ = *(s + 5);
	*t++ = *(s + 4);
	*t++ = *(s + 3);
	*t++ = *(s + 2);
	*t++ = *(s + 1);
	*t++ = *(s + 0);
	if ((s[0] & 0xff) == 0x80)
		v = 0.0;
	else
		v = (double)d * 1e-3;
	data[k].e[i] = v;
	if (xflg)
		printf("%d %g kWh\n", i, v);
	s += 8;
  }
  return 0;
}

int GetCurrent(id, ch, ab)
{
  int i, j, k, adr;
  unsigned short d[9];
  char *s;
  double v[9];

  adr = 0x0000 + (ch - 1) * 0x30+ ab * 0x18;

  k = (id - 1) * 12 + (ch - 1) + ab * 6;	// circuittab[] index
  if (xflg)
	printf("ID %d  CH%d-%s (0x%04x) %s\n", id, ch, ab ? "B" : "A", adr,
		circuittab[k].s);


  if (adr < 0x0000 || 0x011C < adr) {
	fprintf(stderr, "GetCurrent(%d, %d, %d): illegal address.\n",
		id, ch, ab);
	return -1;
  }
  Red(id, adr, 24);
  s = buf + 3;
  for (i = 0; i < 9; i++) {
	d[i] = ntohl(*(unsigned long *)s);
	v[i] = (double)d[i] * 1e-2;
	if (xflg) {
		if (*s & 0x80)
			printf("%d -\n", i);
		else
			printf("%d %d %.2f\n", i, d[i], v[i]);
	}
	data[k].i[i] = v[i];
	s += 4;
	if ((i % 3) == 2)
		s += 4;
  }
  return 0;
}

int GetVoltage(id, ch, ab)
{
  int i, j, k, adr;
  unsigned short d[9];
  char *s;
  double v[9];

  adr = 0x0186 + (ch - 1) * 0x48 + ab * 0x24;
  k = (id - 1) * 12 + (ch - 1) + ab * 6;	// circuittab[] index
  if (xflg)
	printf("ID %d  CH%d-%s (0x%04x) %2d %s\n", id, ch, ab ? "B" : "A", adr,
		circuittab[k].n, circuittab[k].s);
  if (adr < 0x0186 || 0x032E < adr) {
	fprintf(stderr, "GetVoltage(%d, %d, %d): illegal address.\n",
		id, ch, ab);
	return -1;
  }
  Red(id, adr, 30);
  s = buf + 3;
  for (i = 0; i < 9; i++) {
	d[i] = ntohl(*(unsigned long *)s);
	v[i] = (double)d[i] * 1e-2;
	if (xflg) {
		if (*s & 0x80)
			printf("%d -\n", i);
		else
			printf("%d %d %g\n", i, d[i], v[i]);
	}
	data[k].v[i] = v[i];
	s += 4;
	if ((i % 3) == 2)
		s += 12;
  }
  return 0;
}

/*
 * ch = 1, 2, .. 6  ab = A: 0, B: 1
 */

int GetFrequency(id, ch, ab)
{
  int i, k, adr;
  unsigned short d[3];
  char *s;
  double v[3];

  adr = 0x0780 + (ch - 1) * 0x0C + ab * 0x06;
  k = (id - 1) * 12 + (ch - 1) + ab * 6;	// circuittab[] index
  if (xflg)
	printf("ID %d  CH%d-%s (0x%04x) %s\n", id, ch, ab ? "B" : "A", adr,
		circuittab[k].s);
  if (adr < 0x0780 || 0x07c6 < adr) {
	fprintf(stderr, "GetFrequency(%d, %d, %d): illegal address.\n",
		id, ch, ab);
	return -1;
  }
  Red(id, adr, 6);	// 12 byte
  s = buf + 3;
  for (i = 0; i < 3; i++) {
	d[i] = ntohl(*(unsigned long *)s);
	v[i] = (double)d[i] * 1e-2;
	s += 4;
	data[k].f[i] = v[i];
	if (xflg)
	printf("%d %d %.2f Hz\n", k, i, data[k].f[i]);
  }
  return 0;
}


int diag(id)
{
  static char *msgtab[] = {
	">0()",
	"<*0.2",
	"=0(̵)",
	"<0()",
	"T(2)ή<3%",
	"R(1)ή<3%",
	"S-T(2-N)Ű<80%",
	"R-S(1-N)Ű<80%",
	"S-T(2-N)Ű<10%",
	"R-S(1-N)Ű<10%",
	NULL
  };
  int i, j, k, w, d;
  char *s;

  Red(id, 0x07D0, 10);
  printf("- ID %d: %s -\n", id, buf + 3);
  Red(id, 0x07E0, 13);
  s = buf + 3;
  w = ntohs(*(unsigned short *)s);
  for (i = 0; i < 12; i++) {
	if (w & (1 << i)) {
		d = ntohs(*(unsigned short *)(s + 2 + 2 * i));
		k = (id - 1) * 12 + i + (i % 2) * 6;	// circuittab[] index
		printf("CH%d%s %2d %s (0x%04x)\n", (i / 2) + 1, (i % 2) ? "B" : "A",
			circuittab[k].n, circuittab[k].s, d);
		for (j = 0; j < 10; j++)
			if (d & (1 << j))
				printf("  %s\n", msgtab[j]);
	}
  }
  return 0;
}

static char *paratab1[] = {
  "CH1 ʬ", "CH2 ʬ", "CH3 ʬ", "CH4 ʬ",
  "Ch5 ʬ", "CH6 ʬ", "Ű()",
  "Ű()", "CH1 ή", "CH2 ή",
  "CH3 ή", "CH4 ή", "CH5 ή",
  "CH6 ή", "CH1 ʰ켡ή", "CH2 ʰ켡ή",
  "CH3 ʰ켡ή", "CH4 ʰ켡ή", "CH5 ʰ켡ή",
  "CH6 ʰ켡ή", "ʰ켡Ű()",
  "ʰ켡Ű()", "CH Ű", "CH Ѿ", "?",
  "ϥå"
};
static char *paratab2[] = {
  "ñꣲ", "ñꣳ", "ꣳ", "?", "ñꣲ", "ñꣲʬ"
};
static char *paratab3[] = {
  "110V", "220V"
};
static char *paratab4[] = {
  "5A", "50A", "100A", "200A", "400A", "600A"
};
static char *paratab5[] = {
  "", "5.00A", "6.00A", "7.50A", "8.00A", "10.0A", "12.0A", "15.0A", "20.0A",
  "25.0A", "30.0A", "40.0A", "50.0A", "60.0A", "75.0A", "80.0A", "100A",
  "120A", "150A", "200A", "250A", "300A", "400A", "500A", "600A",
  "750A", "800A", "1000A", "1200A", "1500A", "1600A", "2000A", "2500A",
  "3000A", "4000A", "5000A", "6000A", "7500A", "8000A", "9000A", "?"
};
static char *paratab6[] = {
  "", "110V", "220V", "440V", "690V", "1100V", "2200V", "3300V", "6600V",
  "11000V", "13200V", "13800V", "15000V", "16500V", "22000V", "24000V",
  "33000V", "66000V", "77000V", "?", "?", "?", "?", "?", "?", "?"
};

/* ѥ᡼ɤߤ */

int RedPar(id)
{
  int i, k, adr, par[24];
  char *s;

  adr = 0x1000;
  Red(id, adr, 24);
  s = buf + 3;
  dump(s, 24 * 2);
  for (i = 0; i < 24; i++) {
	par[i] = ntohs(*(unsigned short *)s);
	s += 2;
  }
  printf("id %d\n", id);
  for (i = 0; i < 24; i++) {
	k = 0;
	if (i == 0)
		printf("# ʬ\n");
	else if (i == 6)
		printf("\n# Ű\n");
	else if (i == 8)
		printf("\n# ή\n");
	else if (i == 14)
		printf("\n# ʰ켡ή\n");
	else if (i == 20)
		printf("\n# ʰ켡Ű\n");
	else if (i == 22)
		printf("\n# CH Ű\n");
	else if (i == 23)
		printf("\n# CH Ѿ\n");
	else
		k = 1;
	if (0 < k)
		printf(" ");
	printf("%02X", par[i]);
  }
  printf("\n");
  return 0;
}


int GetPar(id)
{
  int i, adr, par;
  char *s;

  printf("ID: %d\n", id);
  adr = 0x1000;
  Red(id, adr, 24);
  s = buf + 3;
  /* ʬ */
  for (i = 0; i < 6; i++) {
	par = ntohs(*(short*)s);
	printf("%s(%d) %s\n", paratab1[i], par, paratab2[par]);
	s += 2;
  }
  /* Ű */
  for (i = 0; i < 2; i++) {
	par = ntohs(*(short*)s);
	printf("%s(%d) %s\n", paratab1[6 + i], par, paratab3[par]);
	s += 2;
  }
  /* ή */
  for (i = 0; i < 6; i++) {
	par = ntohs(*(short*)s);
	printf("%s(%d) %s\n", paratab1[8 + i], par, paratab4[par]);
	s += 2;
  }
  /* ʰ켡ή */
  for (i = 0; i < 6; i++) {
	par = ntohs(*(short*)s);
	printf("%s(%02XH) %s\n", paratab1[14 + i], par, paratab5[par]);
	s += 2;
  }
  /* ʰ켡Ű */
  for (i = 0; i < 2; i++) {
	par = ntohs(*(short*)s);
	printf("%s(%02XH) %s\n", paratab1[20 + i], par, paratab6[par]);	
	s += 2;
  }
  /* Ű */
  par = ntohs(*(short*)s);
  printf("%s(%02XH)\n", paratab1[22], par & 0xFF);
  printf("    CH");
  for (i = 0; i < 6; i++)
	printf(" %d", i + 1);
  printf("\n");
  printf("  ");
  for (i = 0; i < 6; i++)
	printf(" %d", ((par >> i) & 0x1) + 1);

  printf("\n");
  s += 2;
  /* ʬѾ */
  par = ntohs(*(short*)s);
  printf("%s(%02XH) O: ʬ, -: ʬ̤\n", paratab1[23], par & 0xFF);
  printf("    CH");
  for (i = 0; i < 12; i++) {
	printf(" %d", i / 2 + 1);
	printf("%s", (i % 2) == 0 ? "A" : "B");
  }
  printf("\n");
  printf("  ");
  for (i = 0; i < 12; i++)
	printf("  %s", ((par >> i) & 0x1) ? "O" : "-" );
  printf("\n");

#ifdef WMS_PE1N
  usleep(50000);
  adr = 0x1020;
  Red(id, adr, 1);
  par = ntohs(*(short*)(buf+3));
  printf("%s %02XH\n", paratab1[adr - 0x1000], par);
#endif
  return 0;
}

/*
 * WMB-AI8 (id = 6)
 */

int getAin(fd, ch, pv) double *pv;
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  id = 6;	/* WMB-DIO8R */
  adr = 0x0000;
  func = 0x04;	/* read analog input (-20.00 %  120.00 %) */
  cnt = 8;
  *buf = id & 0xff;
  *(buf + 1) = func;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  crc = Crc((unsigned char *)buf, 6);
  *(unsigned short *)(buf + 6) = htons(crc & 0xffff);
  if (xflg)
	dump(buf, 8);
  usleep(1820);
  if (write(fd, buf, 8) < 0)
	perror("write()");
  *buf = '\0';
  usleep(1820);
  n = getResponse(fd);
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "getAin(): receive data CRC error.\n");
  n = ntohs(*(short *)(buf + 3 + (ch - 1) * 2));
  if (n >= 65000)
	*pv = 0.0;	// open
  else
	*pv = (double)n * 0.0001 * 10.0;
  return n;
}

/*
 * WMB-DIO8R (id = 5)
 */

int getCount(fd, ch)
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  id = 5;	/* WMB-DIO8R */
  adr = 0x0000 + (ch - 1) * 2;
  func = 0x04;	/* read pulse count */
  cnt = 2;
  *buf = id & 0xff;
  *(buf + 1) = func;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  crc = Crc((unsigned char *)buf, 6);
  *(unsigned short *)(buf + 6) = htons(crc & 0xffff);
  if (xflg)
	dump(buf, 8);
  usleep(1820);
  if (write(fd, buf, 8) < 0)
	perror("write()");
  *buf = '\0';
  usleep(1820);
  n = getResponse(fd);
  if (xflg)
	dump(buf, n);
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "getCount(): receive data CRC error.\n");
  r = ntohl(*(unsigned long *)(buf + 3));
  return r;
}

int getOnoff(fd, ch)
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  id = 5;	/* WMB-DIO8R */
  adr = 0x0000 + (ch - 1);
  func = 0x02;	/* read on/off */
  cnt = 1;
  *buf = id & 0xff;
  *(buf + 1) = func;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  crc = Crc((unsigned char *)buf, 6);
  *(unsigned short *)(buf + 6) = htons(crc & 0xffff);
  if (xflg)
	dump(buf, 8);
  usleep(1820);
  if (write(fd, buf, 8) < 0)
	perror("write()");
  *buf = '\0';
  usleep(1820);
  n = getResponse(fd);
  if (xflg)
	dump(buf, n);
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "getCount(): receive data CRC error.\n");
  r = *(buf + 3) & 0xFF;
  return r;
}

/*
 * reset pulse count ch to val
 */

int resetCounter(fd, ch, val)
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  if (ch < 1 || 8 < ch)
	error("clearCounter(): illegal channel (%d).\n", ch);
  id = 5;	/* WMB-DIO8R */
  adr = 0x0061 + (ch - 1) * 2;
  func = 0x10;	/* reset pulse count */
  cnt = 2;
  *buf = id & 0xff;
  *(buf + 1) = func;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  *(unsigned short *)(buf + 6) = cnt * 2;
  *(unsigned long *)(buf + 7) = htonl(val);
  crc = Crc((unsigned char *)buf, 11);
  *(unsigned short *)(buf + 11) = htons(crc & 0xffff);
  if (xflg)
	dump(buf, 13);
  usleep(1820);
  if (write(fd, buf, 13) < 0)
	perror("write()");
  *buf = '\0';
  n = getResponse2(fd);
//dump(buf, n);
  return 0;
}

/*
 * function code 01H, 02H, 03H, 04H .. read
 */

int getResponse(fd)
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  if ((n = read(fd, buf, 3)) < 0) {
	fprintf(stderr, "reddio(): read() error. (%d)\n", n);
	return -1;
  }
  k = (buf[2] & 0xFF) + 2;
  n += k;
  if ((k = read(fd, buf + 3, k)) < 0) {
	fprintf(stderr, "reddio(): read() error. (%d)\n", n);
	return -1;
  }
  if (buf[1] & 0x80) {
	static char *err[] = {
		"",
		"⡼塼뤬бεǽɤ",
		"⡼塼뤬бΥɥ쥹",
		"ǡ礭"
	};
	printf("Error code 0x%02x (%s)\n", buf[2] & 0xff, err[buf[2] & 0xff]);
	crc =  Crc((unsigned char *)buf, 3);
	if (crc != ntohs(*(unsigned short *)(buf + 3)))
		fprintf(stderr, "response CRC error.\n");
	return -1;

  }
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "response CRC error.\n");
  return n;
}

/*
 * function code 05H, 06H, 08H, 0FH, 10H .. write
 */

int getResponse2(fd)
{
  unsigned short crc;
  unsigned char l, h;
  int id, adr, func, cnt, r, c, n, k;
  char *s;

  if ((n = read(fd, buf, 2)) < 0) {
	fprintf(stderr, "reddio(): read() error. (%d)\n", n);
	return -1;
  }
  if (buf[1] & 0x80) {
	static char *err[] = {
		"",
		"⡼塼뤬бεǽɤ",
		"⡼塼뤬бΥɥ쥹",
		"ǡ礭"
	};
	if ((k = read(fd, buf + 2, 3)) < 0) {
		fprintf(stderr, "reddio(): read() error. (%d)\n", n);
		return -1;
	}
	n += k;
	printf("Error code 0x%02x (%s)\n", buf[2] & 0xff, err[buf[2] & 0xff]);
	crc =  Crc((unsigned char *)buf, 3);
	if (crc != ntohs(*(unsigned short *)(buf + 3)))
		fprintf(stderr, "response CRC error.\n");
	return -1;
  }
  if ((k = read(fd, buf + 2, 6)) < 0) {
	fprintf(stderr, "reddio(): read() error. (%d)\n", n);
	return -1;
  }
  n += k;
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "response CRC error.\n");
  return n;
}

int Red(id, adr, cnt)
{
  unsigned short crc;
  unsigned char l, h;
  int func, r, c, n;
  char *s;

  *buf = id & 0xff;
  if ((id < 5 && adr == 0x07D0) || (5 <= id && adr == 0x2328))
	func = 0x04;
  else if (id == 5)
	func = 0x01;	/* WMB-DIO8R */
  else if (id == 6)
	func = 0x03;	/* WMB-AI8 */
  else
	func = (adr & 0x1000) ? 0x03 : 0x04;
  *(buf + 1) = func;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  *(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
  crc = Crc((unsigned char *)buf, 6);
  *(unsigned short *)(buf + 6) = htons(crc & 0xffff);
  if (xflg)
	dump(buf, 8);
  usleep(1820);
  if (write(fd, buf, 8) < 0)
	perror("write()");
  *buf = '\0';
  usleep(1820);
  n = recv(fd, buf, id);
  crc =  Crc((unsigned char *)buf, n - 2);
  s = buf + n - 2;
  if (crc != ntohs(*(unsigned short *)s))
	fprintf(stderr, "Red(): receive data CRC error.\n");
  r = buf[1] & 0xff;
  c = buf[2] & 0xff;
  if (r & 0x80) {	/* error */
	fprintf(stderr, "Red(): ID(%d) addr(0x%04x) %s\n",
		id, adr, err[c & 0x03]);
	return -1;
  }
  return 0;
}

#if 0
/*
 * ̥ꥻå (ǤդͤǤ)
 * p1 = ͭ()
 * p2 = ͭ()
 * p3 = ̵(٤)
 * p4 = ̵(ſʤ)
 * p5 = ̵(٤)
 * p6 = ̵(ſʤ)
 */

ResetPowerVal(id, ch, ab, p1, p2, p3, p4, p5, p6)
{
  int adr;

  adr = 0x1080 + 0x30 * (ch - 1) + ab * 0x18;
  Wrt(id, adr, 6, p1, p2, p3, p4, p5, p6);

  return 0;
}
#endif

/*
 * Word 񤭹: id: 0x01-0xff, adr: 0x0000-0xffff, cnt: 1-125, ...: word list
 */

int Wrt(int id, int adr, int cnt, ...)
{
  unsigned short crc;
  unsigned char l, h;
  va_list argp;
  int i, p, r, c, n;
  char *t;

  *buf = id & 0xff;
  *(buf + 1) = (cnt == 1) ? 0x06 : 0x10;
  *(unsigned short *)(buf + 2) = htons(adr & 0xffff);
  if (cnt == 1)
	t = buf + 4;
  else {
	*(unsigned short *)(buf + 4) = htons(cnt & 0xffff);
	*(buf + 6) = (cnt << 1) & 0xff;
	t = buf + 7;
  }
  va_start(argp, cnt);
  for (i = 0; i < cnt; i++) {
	p = (int)va_arg(argp, int);
	*(unsigned short *)t = htons(p & 0xffff);
	t += 2;
  }
  va_end(argp);
  crc = Crc((unsigned char *)buf, (cnt == 1) ? 6 : 7 + cnt * 2);
  *(unsigned short *)t = htons(crc & 0xffff);
  t += 2;
  if (xflg) {
	printf("send: ");
	dump(buf, t - buf);
  }
  if ((i = write(fd, buf, t - buf)) < 0)
	perror("write()");
  *buf = '\0';
  n = recv(fd, buf, id);
 /* ۾ 5 byte */
  r = buf[1] & 0xff;
  c = buf[2] & 0xff;
  if (r & 0x80) {	/* error */
	fprintf(stderr, "Wrt(): %s\n", err[c & 0x03]);
	return -1;
  }
  crc =  Crc((unsigned char *)buf, n - 2);
  if (crc != ntohs(*(unsigned short *)(buf + n - 2) & 0xffff))
	fprintf(stderr, "Wrt(): receive data CRC error.\n");
  return 0;
}

int recv(fd, buf, id) char *buf;
{
  int c, c1, c2, func, i, j, lrc, n, siz, len;
  char *s, *t;

  usleep(182000);
  alarm(2);
  s = buf;
  /* id */
  if ((n = read(fd, s, 1)) != 1) {
	fprintf(stderr, "recv() error. n(%d)\n", n);
	return -1;
  }
  /* ǽ 0xF0 ФƤ礬 */
  if ((buf[0] & 0xFF) != id) {
	fprintf(stderr, "* first byte (%02X), id(%02X)\n", buf[0] & 0xFF, id);
	if ((n = read(fd, s, 1)) != 1) {
		fprintf(stderr, "recv() error. n(%d)\n", n);
		return -1;
	}
	if ((buf[0] & 0xFF) != id) {
		fprintf(stderr, "first byte (%02X) not match id(%02X)\n",
			buf[0] & 0xFF, id);
		return -1;
	}
  }
  s++;
  /* function */
  if ((n = read(fd, s, 1)) != 1) {
	fprintf(stderr, "recv() error. n(%d)\n", n);
	return -1;
  }
  func =  buf[1] & 0xFF;
  s++;
  if (buf[1] & 0x80) {
	siz = 3;
	len = 5;
  }
  else if (func == 6) {
	siz = 6;
	len = 8;
  }
  else if (func == 3 || func == 4) {	/* func == 0x10 */
	/* size */
	if ((n = read(fd, s, 1)) != 1) {
		fprintf(stderr, "recv() error. n(%d)\n", n);
		return -1;
	}
	s++;
	siz = (buf[2] & 0xFF) + 2;
	len = 3 + siz;
  }
  else if (func == 0x10) {	/* func == 0x10 */
	siz = 6;
	len = 8;
  }
  else
	error("function (0x%02x) is not supported.", (int)func);
  for (i = 0; i < siz; ) {
	if ((n = read(fd, s, siz)) == 0) {
		usleep(100000);
		continue;
	}
	else if (n < 0) {
		error("ERR - ŸäƤޤ\n", 0);
		exit(1);
	}
	i += n;
	s += n;
  }
  alarm(0);
  n = len;
  if (xflg) {
	printf("recv(): ");
	dump(buf, n);
  }
  return n;
}



unsigned short
Lrc(s, n) unsigned char *s;
{
  int i;
  unsigned x, lrc;

  for (lrc = i = 0; i < n; i += 2) {
	sscanf((char *)s + i, "%2X", &x);
	lrc += x;
  }
  lrc = ~(lrc - 1);
  return lrc & 0xff;
}

unsigned short
Crc(s, n) unsigned char *s;
{
  unsigned short crc = 0xffff;
  unsigned short next;
  unsigned short carry;
  unsigned short i;
  unsigned short crcl;

  while (n--) {
	next = (unsigned short)*s;
	crc ^= next;
	for (i = 0; i < 8; i++) {
		carry = crc & 1;
		crc >>= 1;
		if (carry)
			crc ^= 0xA001;
	}
	s++;
  }
  crcl = (crc & 0xff00) >> 8;
  crc <<= 8;
  crc |= crcl;
  return crc;
}

int Energy()
{
  int i, j;

  printf("                             (kWh)\n");
  printf("    ----- ͭ -----  ------------------ ̵ ----------------\n");
  printf("No.                   ٤    ſʤ    ٤    ſʤ\n");
  for (i = 1; i <= 4; i++)
	for (j = 1; j <= 6; j++) {
		GetEnergy(i, j, 0);
		if (1 < i)
			GetEnergy(i, j, 1);
	}
  for (i = 0; i <= 34; i++) {
	if (circuittab[i].n < 0)
		continue;
	printf("%2d", circuittab[i].n);
	for (j = 0; j < 6; j++)
		printf("%12.2f", data[i].e[j]);
	printf(" %s\n", circuittab[i].s);
  }
  return 0;
}

int Power()
{
  int i, j;

  printf("                                    (W)\n");
  printf("   -------- ֻ ------- -------- Ǿ ------- --------  -------\n");
  printf("No ͭ ̵  Ψ ͭ ̵  Ψ ͭ ̵  Ψ\n");
  for (i = 1; i <= 4; i++)
	for (j = 1; j <= 6; j++) {
		GetPower(i, j, 0);
		if (1 < i)
			GetPower(i, j, 1);
	}
  for (i = 0; i <= 34; i++) {
	if (circuittab[i].n < 0)
		continue;
	printf("%2d", circuittab[i].n);
	for (j = 0; j < 9; j++)
		printf(" %7.2f", data[i].p[j]);
	printf(" %s\n", circuittab[i].s);
  }
  return 0;
}

int Voltage()
{
  int i, j;

  printf("                          Ű (V)\n");
  printf("   ------ ֻ ------  ----- Ǿ ------  -----  ------\n");
  printf("No.  1-N    2-N    1-2    1-N    2-N    1-2    1-N    2-N    1-2\n");
  for (i = 1; i <= 4; i++)
	for (j = 1; j <= 6; j++) {
		GetVoltage(i, j, 0);
		if (1 < i)
			GetVoltage(i, j, 1);
	}
  for (i = 0; i <= 34; i++) {
	if (circuittab[i].n < 0)
		continue;
	printf("%2d", circuittab[i].n);
	for (j = 0; j < 9; j++)
		printf("%7.2f", data[i].v[j]);
	printf(" %s\n", circuittab[i].s);
  }
  return 0;
}

int Current()
{
  int i, j;

  printf("                         ή (A)\n");
  printf("   ----- ֻ ---- ----- Ǿ ---- -----  ---\n");
  printf("No.   1     2     N     1     2     N     1     2     N\n");
  for (i = 1; i <= 4; i++)
	for (j = 1; j <= 6; j++) {
		GetCurrent(i, j, 0);

		if (1 < i)
			GetCurrent(i, j, 1);
	}
  for (i = 0; i <= 34; i++) {
	if (circuittab[i].n < 0)
		continue;
	printf("%2d", circuittab[i].n);
	for (j = 0; j < 9; j++)
		printf("%6.2f", data[i].i[j]);
	printf(" %s\n", circuittab[i].s);
  }
  return 0;
}

int Frequency()
{
  int i, j;

  printf("        ȿ (Hz)\n");
  printf("No. ֻ Ǿ \n");
  for (i = 1; i <= 4; i++)
	for (j = 1; j <= 6; j++) {
		GetFrequency(i, j, 0);
		if (1 < i)
			GetFrequency(i, j, 1);
	}
  for (i = 0; i <= 34; i++) {
	if (circuittab[i].n < 0)
		continue;
	printf("%2d ", circuittab[i].n);
	for (j = 0; j < 3; j++)
		printf("%7.2f", data[i].f[j]);
	printf(" %s\n", circuittab[i].s);
  }
  return 0;
}

/*
 * ͡Ǿͥꥻå
 */

int ResetMinMax(argc, argv) char **argv;
{
  int adr, i, n, id, ch, ab;

  for (i = 0; i < argc; i++) {
	if (strcmp(argv[i], "all") == 0) {
		for (id = 1; id <= 4; id++) {
			adr = 0x1034;
			Wrt(id, adr, 1, 0);
		}
		continue;
	}
	n = -1;
	sscanf(argv[i], "%d", &n);
	if (Circuit2IdCh(n, &id, &ch, &ab) < 0)
		continue;
	adr = 0x1028 + (ch - 1) * 2 + ab;
	Wrt(id, adr, 1, 0);
  }
  return 0;
}

/*
 * ̥ꥻå
 */

int ResetEnergy(argc, argv) char **argv;
{
  int adr, i, n, id, ch, ab;

  for (i = 0; i < argc; i++) {
	if (strcmp(argv[i], "all") == 0) {
		for (id = 1; id <= 4; id++) {
			adr = 0x107F;
			Wrt(id, adr, 1, 0);
		}
		continue;
	}

	n = -1;
	sscanf(argv[i], "%d", &n);
	if (Circuit2IdCh(n, &id, &ch, &ab) < 0)
		continue;
	adr = 0x1080 + (ch - 1) * 0x30 + ab * 0x18;
	Wrt(id, adr, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0);
  }
  return 0;
}

int Circuit2IdCh(n, pid, pch, pab) int *pid, *pch, *pab;
{
  int i, j, k, id, ch, ab;

  for (j = 0; j < 48; j++)
	if (circuittab[j].n == n)
		break;
  if (j == 48) {
	fprintf(stderr, "ϩֹ %d ¸ߤޤ\n", n);
	return -1;
  }
  id = j / 12 + 1;
  k = j % 12;
  if (5 < k) {
	ab = 1;
	k -= 6;
  }
  else
	ab = 0;
  ch = k % 6 + 1;
  *pid = id;
  *pch = ch;
  *pab = ab;
  return 0;
}

int dump(s, n) char *s;
{
  int i;

  for (i = 0; i < n; i++) {
	if (0 < i)
		printf(" ");
	printf("%02X", s[i] & 0xFF);
  }
  printf("\n");
  return 0;
}

void error(s, t) char *s; int t;
{
  fprintf(stderr, s, t);
  fprintf(stderr, "\n");
  exit(1);
}

/*
 * check module type
 */

int Module()
{
  int id, adr;

  printf("Adr Module\n");
  for (id = 1; id <= 4; id++) {
	Red(id, 0x07D0, 10);
	if (xflg)
		dump(buf, 128);
	printf("%2d %s\n", id, buf + 3);	/* get module tyle letter */
  }
  for (id = 5; id <= 6; id++) {
	Red(5, 0x2328, 64);
	if (xflg)
		dump(buf, 128);
	printf("%2d %s:%s:%s:%s:%s:%s\n", id,
		buf+11,buf+43,buf+75,buf+83,buf+91,buf+99);
  }
  return 0;
}

void Usage()
{
  static char *t[] = {
"ޥ	%s - ŵ() MODBUS ⥸塼",
"ʸ		%s [ץ]",
"ץ	-C ϩ []	# 󥿡ꥻå",
"		-I		# ŷ (lx)",
"		-M ϩ ..	# ͡Ǿͥꥻå",
"		-R ϩ ..	# ̥ꥻå",
"		-S		# ѥ᡼",
"		-a		# ʥŰɽ",
"		-c		# ɤ߼",
"		-d		# ",
"		-e		# ɽ",
"		-f		# ȿɽ",
"		-i		# ήɽ",
"		-m		# ⥸塼ǧ",
"		-p		# ɽ",
"		-v		# Űɽ",
"		-s		# ѥ᡼ɽ",
"		-x		# ǥХåɽ",
"ֲϩפ 1  34 ޤǤʬֲϩֹꤷޤallפȵҤ",
"ϩˤʤޤ",
"-c Τߤꤷɤ߼ϡ֥ѥ륹 ON/OFF(1: ON, 0: OFF)",
"ɽޤ",
"-W ꤹȡֿƻή(L) ή(m^3)פɽޤ",
  NULL
  };
  int i;

  for (i = 0; t[i] != NULL; i++) {
	fprintf(stderr, t[i], cmd);
	fprintf(stderr, "\n");
  }
  exit(1);
}
